/*
 * Copyright (c) 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004, 2005, 2006,
 *               2008, 2009, 2010, 2011, 2012, 2013, 2014
 *      Inferno Nettverk A/S, Norway.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. The above copyright notice, this list of conditions and the following
 *    disclaimer must appear in all copies of the software, derivative works
 *    or modified versions, and any portions thereof, aswell as in all
 *    supporting documentation.
 * 2. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *      This product includes software developed by
 *      Inferno Nettverk A/S, Norway.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * Inferno Nettverk A/S requests users of this software to return to
 *
 *  Software Distribution Coordinator  or  sdc@inet.no
 *  Inferno Nettverk A/S
 *  Oslo Research Park
 *  Gaustadallen 21
 *  NO-0349 Oslo
 *  Norway
 *
 * any improvements or extensions that they make and grant Inferno Nettverk A/S
 * the rights to redistribute these changes.
 *
 */

#include "common.h"
#include "config_parse.h"

static const char rcsid[] =
"$Id: barefootd_udp.c,v 1.85 2014/01/27 14:38:39 karls Exp $";


extern iostate_t iostate;
extern int       rawsocket;

static sockshost_t *
getudptarget(const char *buf,
          const size_t buflen,
          const struct sockaddr_storage *from,
          const int s,
          const connectionstate_t *state,
          const int dstc,
          udptarget_t *dstv,
          const struct sockaddr_storage *to,
          rule_t *rule,
          sockshost_t *targethost,
          char *emsg,
          const size_t emsglen);
/*
 * Gets the target address for a packet received from "from" on the
 * local address "to".
 * In Barefoot it will normally not vary since the bounce-to address
 * is set in the rule and fixed there.  A sighup can change the
 * bounce-to address however, so we must be prepared to handle a change
 * of destination in Barefoot's case too.
 *
 * "rule" is the ruleinfo we have for this packet so far.  If no error
 * occurs, "rule" will upon return  contain the rule matched for this packet.
 *
 * Returns "host" on success, filled in appropriately, or NULL on failure.
 * On failure, "emsg" of "emsglen" may contain an error string describing
 * the reason for failure.
 */

static int
io_allocateresources(rule_t *rule, const struct sockaddr_storage *client,
                     char *emsg, const size_t emsglen);
/*
 * (Re)allocates resources for client "client" as given by "rule".
 * Assumes that if the shared memory pointers in "rule" are not NULL,
 * the resources have been allocated already.
 *
 * Returns 0 on on success, -1 on failure.  On failure, the reason is
 * written to "emsg", which must be of at least size "emsglen".
 */


iostatus_t
io_udp_client2target(client, manytargets, cauth, state,
                     clog, dlog, badfd, packetrule, bwused)
   sockd_io_direction_t *client;
   sockd_io_direction_t *manytargets;
   const authmethod_t *cauth;
   connectionstate_t *state;
   iologaddr_t *clog;
   iologaddr_t *dlog;
   int *badfd;
   rule_t *packetrule;
   size_t *bwused;
{
   /*
    * We are called because a packet has received on the client-socket
    * client->s.  Since this socket is shared by potentially many
    * clients, we don't know which of the target objects in
    * "manytargets->dstv" this client relates to.  If it is a new client,
    * it will not yet have a object in manytargets, and we will have to
    * allocate one for it, assuming the client is allowed.  If it is not
    * allowed, we don't waste resources allocating a target object for it.
    *
    * If the packet is from an existing client, we will be able
    * to reuse the rule verdict from last time (also saved in the
    * target object belonging to this client).  Since the only clients we
    * save are those that are permitted, the rule will always be a pass rule.
    */
   const char *function = "io_udp_client2target()";
   iostatus_t iostatus = IO_NOERROR;
   udptarget_t *target = NULL;
   recvfrom_info_t recvflags;
   sendto_info_t sendtoflags;
   struct sockaddr_storage targetaddr;
   socklen_t len;
   ssize_t w, emsglen;
   ssize_t packetlen;
   char hosta[MAXSOCKSHOSTSTRING], hostb[MAXSOCKSHOSTSTRING], emsg[2048],
        packet[SOCKD_BUFSIZE], tmpbuf[2048];
   int permit = 0,
       icmp_type = ICMP_TYPE_UNREACHABLE,
       icmp_code = ICMP_CODE_UNREACHABLE_PORT;

   /*
    * Don't know what the target socket will be until we've read the packet
    * and figured out what client it belongs to, if it is not a new client.
    *
    * Set it to -1 until then to indicate that errors occurring on the target
    * side before we've figured things out don't belong to any specific
    * target.
    */
   manytargets->s = -1;

   *emsg = NUL;

   recvflags.side = INTERNALIF;
   recvflags.peer = client->raddr;
   recvflags.type = SOCK_DGRAM;

   len = sizeof(client->raddr);
   if ((packetlen = socks_recvfrom(client->s,
                                   packet,
                                   sizeof(packet),
                                   0,
                                   &client->raddr,
                                   &len,
                                   &recvflags,
                                   &client->auth)) == -1) {
      if (ERRNOISPREVIOUSPACKET(errno)) {
         /*
          * error is from a previous packet sent by us out on this socket,
          * i.e., from manytargets to client.
          * Note that Linux apparently can return this error even if the
          * socket is not connected.
          *
          * Should assume the manytargets has closed and we should end the
          * session for now, but we don't know what client the session
          * belongs to. :-/
          */

         *badfd = manytargets->s;

         if (rawsocket == -1) {
            /*
             * Barefoot can have many udp clients and we can't know here
             * which one this error is for.  If we have a raw socket, it
             * should be picked up via the raw socket with more detailed
             * information and logged there.  Since we don't have a raw
             * socket however, do try to log at least something.
             */

            emsglen = snprintf(emsg, sizeof(emsg),
                               "%s (delayed error)", strerror(errno));

            iolog(packetrule,
                  state,
                  OPERATION_TMPERROR,
                  dlog,
                  clog,
                  NULL,
                  NULL,
                  emsg,
                  emsglen);
         }

         return IO_TMPERROR;
      }

      *badfd = client->s;

      if (ERRNOISTMP(errno))
         return IO_EAGAIN;
      else {
         log_unexpected_udprecv_error(function, client->s, errno, INTERNALIF);
         return IO_IOERROR;
      }
   }

   if ((iostatus = io_packet_received(&recvflags,
                                      packetlen,
                                      &client->raddr,
                                      &client->laddr)) != IO_NOERROR) {

      *badfd = client->s;
      return iostatus;
   }

   /*
    * Read a packet ok.  Now figure out if and where it should be forwarded.
    */

   slog(LOG_DEBUG, "%s: received udp packet from %s, packetlength: %ld",
        function, sockaddr2string(&client->raddr, NULL, 0), (long)packetlen);

   sockaddr2sockshost(&client->raddr, &client->host);
   clog->peer       = client->host;
   clog->peer_isset = 1;

   target = clientofclientaddr(&client->raddr,
                               manytargets->dstc,
                               manytargets->dstv);

   slog(LOG_DEBUG, "%s: packet from %s to %s, %s client",
        function,
        sockaddr2string(&client->raddr, NULL, 0),
        target == NULL ?
         "<not yet determined>" : sockshost2string(&target->raddrhost, NULL, 0),
        target == NULL ? "a new" : "an existing");


   SASSERTX(iostatus == IO_NOERROR);

   if (target != NULL) {
      /*
       * Not a new client.  We can reuse the previously saved rule and
       * save ourselves a large amount of work.
       */

      SASSERTX(target->s != -1);
      manytargets->s = target->s; /* update as soon as known. */

      /* we don't waste resources maintaining sessions for blocked clients. */
      SASSERTX(target->crule.verdict == VERDICT_PASS);

      /* so caller gets update regarding what rule we used. */
      /* XXX needed for what? */
      *packetrule       = target->crule;

      dlog->peer        = target->raddrhost;
      dlog->peer_isset  = 1;

      slog(LOG_DEBUG,
           "%s: UDP packet #%"PRIu64" from existing client %s, for target %s.  "
           "Already set up with %s #%lu (%s), so reusing that",
           function,
           target->client_read.packets + 1,
           sockaddr2string(&client->raddr, NULL, 0),
           sockshost2string(&target->raddrhost, NULL, 0),
           objecttype2string(packetrule->type),
           (unsigned long)packetrule->number,
           verdict2string(packetrule->verdict));

      permit = (packetrule->verdict == VERDICT_PASS);
      SASSERTX(permit);

      sockaddr2sockshost(&target->laddr, &dlog->local);
      dlog->local_isset = 1;

      /*
       * possibly target->raddrhost resolves to something different now,
       * but it's probably better to keep the originally resolved name
       * for all packets from this clinet as we don't know how the overlaying
       * protocol will be affected it the same client sends packet to changing
       * target server addresses.  E.g. if it's a DNS-loadsharing thing and
       * the target servers are maintaining some state for clients, it
       * could become problematic unless we keep using the same target
       * ipaddress for this client.
       */
      dlog->peer        = target->raddrhost;
      dlog->peer_isset  = 1;
   }
   else {
      /*
       * No such luck.  Must do a rulelookup on this packet to see
       * if it is permitted.
       */
      sockshost_t host;

      if (getudptarget(packet,
                       (size_t)packetlen,
                       &client->raddr,
                       client->s,
                       state,
                       manytargets->dstc,
                       manytargets->dstv,
                       &client->laddr,
                       packetrule,
                       &host,
                       emsg,
                       sizeof(emsg)) == NULL) {
          slog(LOG_DEBUG,
               "%s: failed to get target of packet from %s, received on %s: %s",
               function,
               sockaddr2string(&client->raddr, hosta, sizeof(hosta)),
               sockaddr2string(&client->laddr, hostb, sizeof(hostb)),
               emsg);

          iostatus  = IO_TMPERROR;
          icmp_code = ICMP_CODE_UNREACHABLE_DESTHOSTUNKNOWN;
      }
      else
         SASSERTX(host.atype != SOCKS_ADDR_NOTSET);

      if (host.atype != SOCKS_ADDR_NOTSET) {
         dlog->peer       = host;
         dlog->peer_isset = 1;
      }
      else
         dlog->peer_isset = 0;

      permit = (packetrule->verdict == VERDICT_PASS);

      if (permit) {
         int gaierr;

         sockshost2sockaddr2(&host, &targetaddr, &gaierr, emsg, sizeof(emsg));
         if (!IPADDRISBOUND(&targetaddr)) {
            if (gaierr != 0) {
               SASSERTX(host.atype == SOCKS_ADDR_DOMAIN);
               log_resolvefailed(host.addr.domain, EXTERNALIF, gaierr);
            }

            iostatus  = IO_TMPERROR;
            icmp_code = ICMP_CODE_UNREACHABLE_DESTHOSTUNKNOWN;
         }
      }
      else
         iostatus = IO_BLOCK;

      if (!permit || iostatus != IO_NOERROR) {
         iolog(packetrule,
               state,
               IOOP(!permit, iostatus),
               clog,
               dlog,
               NULL,
               NULL,
               emsg,
               strlen(emsg));

         if (permit) {
            /*
             * Packetforwarding was permitted, but something else failed.
             */
            SASSERTX(iostatus != IO_NOERROR);
            SASSERTX(IPADDRISBOUND(&targetaddr));

            /*
             * we only send icmp-errors to permitted clients.
             */
            send_icmperror(rawsocket,
                           &client->laddr,
                           &client->raddr,
                           &targetaddr,
                           -1,
                           -1,
                           icmp_type,
                           icmp_code);

         }

         return iostatus;
      }

      SASSERTX(host.atype != SOCKS_ADDR_NOTSET);
      SASSERTX(iostatus   == IO_NOERROR);
      SASSERTX(permit);

      slog(LOG_DEBUG,
           "%s: udp packet #%"PRIu64" from client %s received on %s for dst %s",
           function,
           target == NULL ? 1 : target->client_read.packets,
           sockaddr2string(&client->raddr, NULL, 0),
           sockaddr2string(&client->laddr, hosta, sizeof(hosta)),
           sockshost2string(&host, hostb, sizeof(hostb)));

      /*
       * The packet is ok and from an acceptable client address.
       * Now see if we can allocate the necessary resources for this client.
       */

      SASSERTX(packetrule->bw == NULL);
      SASSERTX(packetrule->ss == NULL);

      if (io_allocateresources(packetrule,
                               &client->raddr,
                               emsg,
                               sizeof(emsg)) == 0) {
         udptarget_t newclient;

         if (initclient(-1,
                        &client->raddr,
                        &host,
                        &targetaddr,
                        packetrule,
                        emsg,
                        sizeof(emsg),
                        &newclient) != NULL) {
            if ((target = addclient(&client->laddr,
                                    &newclient,
                                    &manytargets->dstc,
                                    &manytargets->dstcmax,
                                    &manytargets->dstv,
                                    state,
                                    packetrule)) == NULL) {
               snprintf(emsg, sizeof(emsg), "could not add client %s: %s",
                        sockaddr2string(&client->raddr, NULL, 0),
                        strerror(errno));

               iostatus = IO_TMPERROR;
            }
            else {
               manytargets->s = target->s; /* update as soon as known. */

               /*
                * so we can reuse it next time we get a packet from this client.
                */
               target->crule = *packetrule;
            }
         }
         else
            iostatus = IO_TMPERROR;
      }
      else {
         /* no resources allocated, no resources to free. */
         SHMEM_CLEAR(packetrule, SHMEM_ALL, 1);

         iostatus            = IO_BLOCK;
         packetrule->verdict = VERDICT_BLOCK;
         permit              = 0;
      }

      if (permit && iostatus == IO_NOERROR) {
         SASSERTX(target != NULL);

         if (redirect(target->s,
                      &target->laddr,
                      state->command,
                      &packetrule->rdr_from,
                      tmpbuf,
                      sizeof(tmpbuf)) != 0) {
            snprintf(emsg, sizeof(emsg), "redirect failed: %s", tmpbuf);
            iostatus = IO_TMPERROR; /* XXX more correct error code. */
         }

         sockaddr2sockshost(&target->laddr, &dlog->local);
         dlog->local_isset = 1;
      }

      if (iostatus != IO_NOERROR || !permit) {
         iolog(packetrule,
               state,
               IOOP(!permit, iostatus),
               clog,
               dlog,
               NULL,
               NULL,
               emsg,
               strlen(emsg));

         send_icmperror(rawsocket,
                        &client->laddr,
                        &client->raddr,
                        &targetaddr,
                        -1,
                        -1,
                        icmp_type,
                        iostatus == IO_NOERROR ?
                           ICMP_CODE_UNREACHABLE_HOSTPROHIBITED : icmp_code);

         if (iostatus == IO_NOERROR) {
            SASSERTX(!permit);

            return IO_TMPBLOCK;
         }
         else
            return iostatus;
      }

      SASSERTX(iostatus == IO_NOERROR);
      SASSERTX(permit);
      SASSERTX(target != NULL);

      setconfsockoptions(target->s,
                         client->s,
                         SOCKS_UDP,
                         0,
                         packetrule->socketoptionc,
                         packetrule->socketoptionv,
                         SOCKETOPT_PRE | SOCKETOPT_ANYTIME,
                         SOCKETOPT_PRE | SOCKETOPT_ANYTIME);

      /*
       * Should we connect to the destination also?  If the client will only
       * be sending udp packets to one address, connecting should be more
       * efficient.
       */
      if (sockscf.udpconnectdst) {
         slog(LOG_DEBUG, "%s: connecting fd %d to %s on behalf of client %s",
              function,
              target->s,
              sockshost2string(&target->raddrhost, hosta, sizeof(hosta)),
              sockshost2string(&client->host, hostb, sizeof(hostb)));

         if (socks_connecthost(target->s,
                               EXTERNALIF,
                               &target->raddrhost,
                               NULL,
                               NULL,
                               (long)-1,
                               emsg,
                               sizeof(emsg)) != 0) {
            struct sockaddr_storage faddr;
            socklen_t flen = sizeof(faddr);

            if (getsockname(target->s, TOSA(&faddr), &flen) == 0)
               sockaddr2string(&faddr, hosta, sizeof(hosta));
            else
               STRCPY_ASSERTSIZE(hosta, "N/A");

            swarn("%s: could not connect to target address %s, given in "
                  "rule #%lu, from local address %s.  This might be "
                  "configuration error, either in the kernel routing table, "
                  "or in our %s %s file",
                  function,
                  sockshost2string(&target->raddrhost, NULL, 0),
                  (unsigned long)packetrule->number,
                  hosta,
                  PRODUCT,
                  sockscf.option.configfile);

            iolog(packetrule,
                  state,
                  OPERATION_ERROR,
                  clog,
                  dlog,
                  NULL,
                  NULL,
                  emsg,
                  strlen(emsg));

            iostatus = IO_IOERROR;
            *badfd   = target->s;

            return iostatus;
         }

         target->isconnected = 1;
      }
   }

   SASSERTX(target != NULL);
   SASSERTX(iostatus == IO_NOERROR);
   SASSERTX(permit);

   target->client_read.bytes   += recvflags.fromsocket;
   target->client_read.packets += 1;

   /*
    * Also client so caller can know how much i/o was done on this call,
    * as he otherwise would have no idea, not knowing what the client was.
    */
   client->read.bytes   += recvflags.fromsocket;
   client->read.packets += 1;

   sendtoflags.side = EXTERNALIF;

   w = socks_sendto(target->s,
                    packet,
                    packetlen,
                    0,
                    target->isconnected ?  NULL : &target->raddr,
                    target->isconnected ?
                        (socklen_t)0 : (socklen_t)sizeof(target->raddr),
                    &sendtoflags,
                    NULL);

   if (w >= 0) {
      iostatus = io_packet_sent(packetlen,
                                w,
                                &recvflags.ts,
                                &client->raddr,
                                &target->raddr,
                                emsg,
                                sizeof(emsg));

      gettimeofday_monotonic(&target->lastio);

      *bwused = w;

      target->target_written.bytes   += packetlen;
      target->target_written.packets += 1;

      /*
       * Also manytargets so caller knows how much i/o was done on this call.
       */
      manytargets->read.bytes   += recvflags.fromsocket;
      manytargets->read.packets += 1;
   }
   else {
      snprintf(emsg, sizeof(emsg), "sendto of %lu bytes failed: %s",
               (unsigned long)packetlen, strerror(errno));

      iostatus = IOSTATUS_UDP_SEND_FAILED(errno);
      *badfd   = target->s;
   }

   if (iostatus != IO_NOERROR) {
      iolog(packetrule,
            state,
            IOOP(0, iostatus),
            clog,
            dlog,
            NULL,
            NULL,
            emsg,
            strlen(emsg));

      send_icmperror(rawsocket,
                     &client->laddr,
                     &client->raddr,
                     &target->raddr,
                     -1,
                     -1,
                     icmp_type,
                     iostatus == IO_NOERROR ?
                        ICMP_CODE_UNREACHABLE_HOSTPROHIBITED : icmp_code);

      return iostatus;
   }

   SASSERTX(iostatus == IO_NOERROR);
   SASSERTX(w == packetlen);

   iolog(packetrule,
         state,
         OPERATION_IO,
         clog,
         dlog,
         NULL,
         NULL,
         packet,
         packetlen);

   return IO_NOERROR;
}

iostatus_t
io_udp_target2client(client, manytargets, cauth, state,
                     clog, dlog, badfd, packetrule, bwused)
   sockd_io_direction_t *client;
   sockd_io_direction_t *manytargets;
   const authmethod_t *cauth;
   connectionstate_t *state;
   iologaddr_t *clog;
   iologaddr_t *dlog;
   int *badfd;
   rule_t *packetrule;
   size_t *bwused;
{
   /*
    * In this case a packet has been received on one of the manytargets sockets
    * set up.  Each manytargets socket must have a matching client, so when
    * we are called, both the client and manytargets objects we are passed
    * should be set up correctly.
    *
    * Since we don't save state for blocked clients, the verdict must be
    * permit and there is little more to do than pass the packet through.
    */
   const char *function = "io_udp_target2client()";
   iostatus_t iostatus = IO_NOERROR;
   udptarget_t *target = NULL;
   recvfrom_info_t recvflags;
   sendto_info_t sendtoflags;
   socklen_t len;
   ssize_t w, payloadlen;
   size_t emsglen;
   char hosta[MAXSOCKSHOSTSTRING], hostb[MAXSOCKSHOSTSTRING], emsg[1024],
        payload[SOCKD_BUFSIZE];
   int permit = 0,
       icmp_type = ICMP_TYPE_UNREACHABLE,
       icmp_code = ICMP_CODE_UNREACHABLE_PORT;

   SASSERTX(manytargets->s != -1);

   target = clientofsocket(manytargets->s,
                           manytargets->dstc,
                           manytargets->dstv);
   SASSERTX(target != NULL);

   permit = (packetrule->verdict == VERDICT_PASS);
   SASSERTX(permit); /* if not permitted, why would we keep this session? */

   *emsg = NUL;

   recvflags.side = EXTERNALIF;
   recvflags.peer = manytargets->raddr;
   recvflags.type = SOCK_DGRAM;

   len = sizeof(manytargets->raddr);
   if ((payloadlen = socks_recvfrom(manytargets->s,
                                    payload,
                                    sizeof(payload),
                                    0,
                                    &manytargets->raddr,
                                    &len,
                                    &recvflags,
                                    &manytargets->auth)) == -1) {
      if (ERRNOISPREVIOUSPACKET(errno)) {
         /*
          * error is from the manytargets of an earlier packet from client,
          * sent by us out on this socket (i.e., it refers to a packet sent
          * by the client.  Note that Linux apparently can return this error
          * even if the socket is not connected.
          * Don't treat it as fatal, more packets could come, and those
          * may be accepted by the manytargets.
          */

         emsglen = snprintf(emsg, sizeof(emsg),
                            "%s (delayed error)", strerror(errno));

         *badfd = client->s;

         if (rawsocket == -1) {
            /*
             * No raw socket and error is from manytargets.  Can not be sure,
             * but assume the target is the same as before (i.e., no sighup
             * has changed it).
             */
            const int oldcommand = state->command; /* udpreply presumably. */

            state->command = SOCKS_UDPASSOCIATE;
            iolog(packetrule,
                  state,
                  BAREFOOTD ? OPERATION_ERROR : OPERATION_TMPERROR,
                  clog,
                  dlog,
                  NULL,
                  NULL,
                  emsg,
                  emsglen);
            state->command = oldcommand;
         }
         else {
            /*
             * have a rawsocket, so can try to send an icmp error.
             * Also assume the function that reads from the rawsocket
             * will log the error more accurately than we can here, so
             * don't try.
             */
            send_icmperror(rawsocket,
                           &client->laddr,
                           &client->raddr,
                           &manytargets->raddr,
                           -1,
                           -1,
                           icmp_type,
                           icmp_code);
         }

         return IO_IOERROR;
      }

      *badfd = manytargets->s;

      if (ERRNOISTMP(errno))
         return IO_EAGAIN;
      else {
         log_unexpected_udprecv_error(function,
                                      manytargets->s,
                                      errno,
                                      EXTERNALIF);

         return IO_IOERROR;
      }
   }

   iostatus = io_packet_received(&recvflags,
                                 payloadlen,
                                 &manytargets->raddr,
                                 &manytargets->laddr);

   gettimeofday_monotonic(&target->lastio);

   if (iostatus != IO_NOERROR) {
      *badfd = manytargets->s;
      return iostatus;
   }


   target->target_read.bytes   += recvflags.fromsocket;
   target->target_read.packets += 1;

   /*
    * Also manytargets so caller knows how much i/o was done on this call.
    */
   manytargets->read.bytes   += recvflags.fromsocket;
   manytargets->read.packets += 1;

   slog(LOG_DEBUG,
        "%s: packet #%"PRIu64" from manytargets %s received for client %s "
        "on fd %d.  Packetlength is %ld",
        function,
        manytargets->read.packets,
        sockaddr2string(&manytargets->raddr, hosta, sizeof(hosta)),
        sockaddr2string(&client->raddr, hostb, sizeof(hostb)),
        manytargets->s,
        (long)payloadlen);

   /*
    * All is fine.  Now forward the packet to the client.
    * There is no rulespermit() here as we have no socks-rules to
    * check.  Instead we depend on the admin configuring whether we
    * should connect to manytargets (default: yes) or not.
    * If we have connected to the manytargets, the kernel should make sure we
    * only receive replies from the manytargets (bounce to address).  If not,
    * it means the admin wants to receive replies from other addresses too.
    * Up to him.
    */

   sendtoflags.side = INTERNALIF;

   w = socks_sendto(client->s,
                    payload,
                    payloadlen,
                    0,
                    client->state.isconnected ?  NULL : &client->raddr,
                    client->state.isconnected ?
                        (socklen_t)0 : (socklen_t)sizeof(client->raddr),
                    &sendtoflags,
                    &client->auth);

   if (w >= 0) {
      iostatus = io_packet_sent(payloadlen,
                                w,
                                &recvflags.ts,
                                &manytargets->raddr,
                                &client->raddr,
                                emsg,
                                sizeof(emsg));

      *bwused = w;

      target->client_written.bytes   += sendtoflags.tosocket;
      target->client_written.packets += 1;

      /*
       * Also client so caller knows how much i/o was done on this call.
       */
      client->written.bytes   += sendtoflags.tosocket;
      client->written.packets += 1;
   }
   else {
      snprintf(emsg, sizeof(emsg), "sendto of %lu bytes failed: %s",
               (unsigned long)payloadlen, strerror(errno));

      iostatus = IOSTATUS_UDP_SEND_FAILED(errno);
      *badfd   = client->s;
   }

   if (iostatus != IO_NOERROR) {
      iolog(packetrule,
            state,
            IOOP(0, iostatus),
            dlog,
            clog,
            NULL,
            NULL,
            emsg,
            strlen(emsg));

      send_icmperror(rawsocket,
                     &manytargets->laddr,
                     &manytargets->raddr,
                     &client->raddr,
                     -1,
                     -1,
                     icmp_type,
                     iostatus == IO_NOERROR ?
                        ICMP_CODE_UNREACHABLE_HOSTPROHIBITED : icmp_code);

      return iostatus;
   }

   SASSERTX(w == (ssize_t)payloadlen);

   iolog(packetrule,
         state,
         OPERATION_IO,
         dlog,
         clog,
         NULL,
         NULL,
         payload,
         payloadlen);

   return IO_NOERROR;
}


sockshost_t *
getudptarget(buf, buflen, from, s, state, dstc, dstv, to, rule, host,
          emsg, emsglen)
   const char *buf;
   const size_t buflen;
   const struct sockaddr_storage *from;
   const int s;
   const connectionstate_t *state;
   const int dstc;
   udptarget_t *dstv;
   const struct sockaddr_storage *to;
   rule_t *rule;
   sockshost_t *host;
   char *emsg;
   const size_t emsglen;
{
   const char *function = "getudptarget()";
   authmethod_t cauth, socksauth;
   sockshost_t fromhost, tohost;
   connectionstate_t rulestate;
   char src[MAXSOCKADDRSTRING], dst[MAXSOCKADDRSTRING];

   /*
    * No socks header.  The dst is the fixed bounce-to address in the
    * matching client-rule.  We will thus only know what the destinations
    * address is after we've found the client-rule and looked up the
    * bounce-to address in that.
    */

#if DIAGNOSTIC
   SASSERTX(clientofclientaddr(from, dstc, dstv) == NULL);
#endif /* DIAGNOSTIC */

   slog(LOG_DEBUG,
        "%s: received packet of length %lu from client %s on %s, using fd %d ",
        function,
        (unsigned long)buflen,
        sockaddr2string(from, src, sizeof(src)),
        sockaddr2string(to, dst, sizeof(dst)),
        s);

   bzero(&cauth, sizeof(cauth));
   bzero(&socksauth, sizeof(socksauth));

   /* no auth is supported for udp in Barefoot. */
   cauth.method = socksauth.method = AUTHMETHOD_NOTSET;

   rulestate         = *state;
   rulestate.command = SOCKS_BOUNCETO;

   (void)rulespermit(s,
                     from,
                     to,
                     &cauth,
                     &socksauth,
                     rule,
                     &rulestate,
                     sockaddr2sockshost(from, &fromhost),
                     sockaddr2sockshost(to,   &tohost),
                     NULL,
                     emsg,
                     emsglen);

   SASSERTX(rule->extra.bounceto.atype != SOCKS_ADDR_NOTSET);
   ruleaddr2sockshost(&rule->extra.bounceto, host, SOCKS_UDP);

   slog(LOG_DEBUG, "%s: target address for packet from %s on fd %d is %s",
        function,
        sockaddr2string(from, NULL, 0),
        s,
        sockshost2string(host, NULL, 0));

   return host;
}

static int
io_allocateresources(rule, client, emsg, emsglen)
   rule_t *rule;
   const struct sockaddr_storage *client;
   char *emsg;
   const size_t emsglen;
{
   const char *function = "io_allocateresources()";
   clientinfo_t cinfo = { .from = *(const struct sockaddr_storage *)client };

   slog(LOG_DEBUG, "%s: checking if we have the needed resources for client %s",
        function, sockaddr2string(client, NULL, 0));

   if (rule->ss_shmid != 0) {
      SASSERTX(rule->ss == NULL);

      if (sockd_shmat(rule, SHMEM_SS) == 0) {
         SASSERTX(rule->ss != NULL);

         if (!session_use(rule->ss, &cinfo, sockscf.shmemfd, emsg, emsglen)) {
            sockd_shmdt(rule, SHMEM_SS);
            SHMEM_CLEAR(rule, SHMEM_ALL, 1);

            return -1;
         }

         slog(LOG_DEBUG, "%s: used ss_shmid %ld, addr %s, clients %ld",
              function,
              rule->ss_shmid,
              sockaddr2string(client, NULL, 0),
              (long)rule->ss->mstate.clients);
      }

      /* won't be needed until we delete the client. */
      sockd_shmdt(rule, SHMEM_SS);
   }

   if (rule->bw_shmid != 0) {
      SASSERTX(rule->bw == NULL);

      if (sockd_shmat(rule, SHMEM_BW) == 0) {
         SASSERTX(rule->bw != NULL);

         bw_use(rule->bw, &cinfo, sockscf.shmemfd);

         slog(LOG_DEBUG, "%s: used bw_shmid %ld, addr %s, clients %ld",
              function,
              rule->bw_shmid,
              sockaddr2string(client, NULL, 0),
              (long)rule->bw->mstate.clients);
      }

      /* keep attached since we will need to update it on every i/o op. */
   }

   return 0;
}

size_t
io_udpclients(ioc, iov, mincount)
   const size_t ioc;
   const sockd_io_t iov[];
   const ssize_t mincount;
{
   const char *function = "io_udpclients()";
   size_t i, allocated;

   for (i = 0, allocated = 0; i < ioc; ++i) {
      if (!iov[i].allocated)
         continue;

      if (iov[i].state.protocol != SOCKS_UDP)
         continue;

      allocated += iov[i].dst.dstc;

      if (mincount != -1 && allocated >= (size_t)mincount)
         break;
   }

   slog(LOG_DEBUG, "%s: have %lu udp client%s",
        function, (unsigned long)allocated, allocated == 1 ? "" : "s");

   return allocated;
}

udptarget_t *
clientofclientaddr(addr, udpclientc, udpclientv)
   const struct sockaddr_storage *addr;
   const size_t udpclientc;
   udptarget_t *udpclientv;
{
   static size_t i;

   if (i < udpclientc
   && sockaddrareeq(addr, &udpclientv[i].client, 0))
      return &udpclientv[i];

   for (i = 0; i < udpclientc; ++i) {
      if (udpclientv[i].s == -1)
         continue;

      if (sockaddrareeq(addr, &udpclientv[i].client, 0))
         return &udpclientv[i];
   }

   return NULL;
}


int
removeclient(s, clientc, clientv)
   const int s;
   size_t *clientc;
   udptarget_t *clientv;
{
   const char *function = "removeclient()";
   char client_str[MAXSOCKADDRSTRING],
        laddr_str[MAXSOCKADDRSTRING],
        raddr_str[MAXSOCKSHOSTSTRING];
   size_t i;

   for (i = 0; i < *clientc; ++i) {
      if (clientv[i].s == s) {
         clientinfo_t cinfo = { .from = clientv[i].client };

         slog(LOG_DEBUG,
              "%s: removing client using fd %d at dstv index %lu.  Rule #%lu, "
              "client address %s, bound %s, dst %s.  "
              "New clientc will become %lu and free fds will become %lu",
              function,
              clientv[i].s,
              (unsigned long)i,
              (unsigned long)clientv[i].crule.number,
              sockaddr2string(&clientv[i].client,
                              client_str,
                              sizeof(client_str)),
              sockaddr2string(&clientv[i].laddr,
                              laddr_str,
                              sizeof(laddr_str)),
              sockaddr2string(&clientv[i].raddr,
                              raddr_str,
                              sizeof(raddr_str)),
              (unsigned long)*clientc - 1,
              (unsigned long)(iostate.freefds + 1));

         close(clientv[i].s);
         ++iostate.freefds;

         SHMEM_UNUSE(&clientv[i].crule,
                     &cinfo,
                     sockscf.shmemfd,
                     SHMEM_ALL);

         memmove(&clientv[i],
                 &clientv[i + 1],
                 sizeof(*clientv) * (*clientc - (i + 1)));
         --(*clientc);

#if DEBUG
         for (i = 0; i < *clientc; ++i)
            slog(LOG_DEBUG,
                 "%s: client using fd %d at dstv index %lu.  Rule #%lu, "
                 "client address %s, bound %s, dst %s",
                 function,
                 clientv[i].s,
                 (unsigned long)i,
                 (unsigned long)clientv[i].crule.number,
                 sockaddr2string(&clientv[i].client,
                                 client_str,
                                 sizeof(client_str)),
                 sockaddr2string(&clientv[i].laddr,
                                 laddr_str,
                                 sizeof(laddr_str)),
                 sockaddr2string(&clientv[i].raddr,
                                 raddr_str,
                                 sizeof(raddr_str)));
#endif /* DEBUG */

         return 0;
      }
   }

   /* NOTREACHED */
   SERRX(0);
}


int
io_remove_session(ioc, iov, laddr, protocol, reason)
   const size_t ioc;
   sockd_io_t iov[];
   const struct sockaddr_storage *laddr;
   const int protocol;
   const iostatus_t reason;
{
   const char *function = "io_remove_session()";
   size_t i;

   SASSERTX(protocol == SOCKS_UDP);

   slog(LOG_DEBUG,
        "%s: searching for local address %s.  Marked for removal, iostatus %d",
        function,
        laddr == NULL ? "<any>" : sockaddr2string(laddr, NULL, 0),
        (int)reason);

   for (i = 0; i < ioc; ++i) {
      if (!iov[i].allocated)
         continue;

      if (iov[i].state.protocol != protocol)
         continue;

      if (laddr != NULL
      &&  memcmp(&iov[i].src.laddr, laddr, sizeof(*laddr)) != 0)
         continue;

      slog(LOG_DEBUG, "removing iov #%lu with %lu active udp session%s",
           (unsigned long)i,
           (unsigned long)iov[i].dst.dstc,
           iov[i].dst.dstc == 1 ? "" : "s");

      while (iov[i].dst.dstc > 0) {
         const size_t clienti = iov[i].dst.dstc;

         /*
          * so io_delete() knows what target-index to sync with to sync.
          */

         iov[i].dst.s = iov[i].dst.dstv[clienti - 1].s;

         io_delete(-1 /* no ack to send */,
                   &iov[i],
                   -1,
                   /* io_delete asserts badfd is not -1 on IO_IOERROR. */
                   reason);

         /*
          * io_delete() will call removeclient(), which will
          * decrement iov[i].dst.dstc.
          */
         SASSERTX(iov[i].dst.dstc < clienti);
      }

      SASSERTX(iov[i].dst.dstc == 0);

      /* io_delete() does not close udp src in barefoot. */
      close(iov[i].src.s);
      socks_freebuffer(iov[i].src.s);

      free(iov[i].dst.dstv);
      bzero(&iov[i], sizeof(iov[i]));

      if (sockscf.state.mother.ack != -1) {
         const char info = SOCKD_FREESLOT_UDP;

         /* ack io slot free. */
         if (socks_sendton(sockscf.state.mother.ack,
                           &info,
                           sizeof(info),
                           sizeof(info),
                           0,
                           NULL,
                           0,
                           NULL,
                           NULL) != sizeof(info))
            slog(sockd_motherexists() ? LOG_WARNING : LOG_DEBUG,
                 "%s: sending ack to mother failed: %s",
                 function, strerror(errno));
      }

      return 0;
   }

   return -1;
}
